/*
 * file.cpp
 *
 *  Created on: Mar 13, 2009
 *      Author: ranton
 */

/*
 * typesupport.cpp
 * This contains basic builtin functions such as type() and len()
 *
 *  Created on: Mar 1, 2009
 *      Author: ranton
 */

#include <cerrno>
#include <cstring>
#include <cstdlib>
#include <cstdio>

// Unix specific IO
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>


#include "meme.h"
#include "memetype.h"
#include "memeutil.h"
#include "memearray.h"
#include "memehash.h"

using namespace std;

static Value bi_open( MemeScope& scope ) {
	const char *path = getVar(scope,"path").toString();
	const char *mode = getVar(scope,"mode").toString();

	int flags = 0;
	if( strchr(mode,'w') )
	{
		if( strchr(mode,'r') )
			flags = O_RDWR;
		else
			flags = O_WRONLY;
		flags += O_CREAT;
	}
	else
		flags = O_RDONLY;

	int fd = open( path, flags, 0777 );
	if( fd < 0 )
	{
		warning("Could not open file %s: %s",path,strerror(errno));
		return Value();
	}
	else
	{
		return Value( fd );
	}
}

static Value bi_close( MemeScope& scope ) {
	int fd = getVar(scope,"fref").toInt();
	if( fd >= 0 )
	{
		if( close(fd) < 0 )
		{
			warning("Error closing file descriptor %d: %s",fd,strerror(errno));
			return Value(false);
		}
		else
			return Value(true);
	}
	return Value(false);
}

static Value bi_read( MemeScope& scope ) {
	int fd = getVarValue(scope,"fref").toInt();
	int count = getVarValue(scope,"count").toInt();
	Value result;
	if( fd >= 0 )
	{
		char *buf = new char[ count ];
		if( !buf )
			error("Out of memory");
		count = read( fd, (char *)buf, count);
		if( count < 0 )
			warning("Error from read: %s",strerror(errno));
		else
			result = Value(valueStr, MemeString::newString( buf, count ));

		delete[] buf;
	}
	return result;
}

static Value bi_write( MemeScope& scope ) {
	int fd = getVar(scope,"fref").toInt();
	MemeString *ms = MemeString::derefString( getVarValue(scope,"str").ref_ );
	if( fd >= 0 )
	{
		int count = write( fd, ms->getText(), ms->getLength() );
		if( count < 0 )
			warning("Error from write: %s", strerror(errno));
		else
			return Value(count);
	}
	return Value();
}

void RegisterFile(MemeScope& scope)
{
	RegisterCppFunc(scope, "open", "path,mode", bi_open);
	// RegisterCppFunc(scope, "ferror", "fref", bi_ferror);
	RegisterCppFunc(scope, "close", "fref", bi_close);
	RegisterCppFunc(scope, "read", "fref,count", bi_read);
	RegisterCppFunc(scope, "write", "fref,str", bi_write);
	// RegisterCppFunc(scope, "freadline", "fref", bi_freadline);

	RegisterConstant(scope, "STDIN",0);
	RegisterConstant(scope, "STDOUT",1);
	RegisterConstant(scope, "STDERR",2);

}
