
#include "memetype.h"
#include "meme.h"
#include <cstdio>

class IntegerPrimitive : public MemePrimitive {
	public:
		// these are not abstract, they all throw
		// errors if called on an object that does
		// not implement
		IntegerPrimitive();
		~IntegerPrimitive();

		// NOTE: this should have variable arguments
		virtual Value ini(const Value& x);

		virtual Value add(const Value& a, const Value& b);
		virtual Value sub(const Value& a, const Value& b);
		virtual Value neg(const Value& x);

		virtual Value mul(const Value& a, const Value& b);
		virtual Value div(const Value& a, const Value& b);
		virtual Value mod(const Value& a, const Value& b);

		virtual Value cmp(const Value& a, const Value& b);

		virtual Value str(const Value& a);
		virtual Value num(const Value& a);
		
};

MemePrimitive *getIntegerPrimitive(void)
{
	return new IntegerPrimitive();
}

// these are not abstract, they all throw
// errors if called on an object that does
// not implement
IntegerPrimitive::IntegerPrimitive()
{
}

IntegerPrimitive::~IntegerPrimitive()
{
}


// NOTE: this should have variable arguments
Value IntegerPrimitive::ini(const Value& x)
{
	return x;
}


Value IntegerPrimitive::add(const Value& a, const Value& b)
{
	return Value( a.i + b.toInt() );
}

Value IntegerPrimitive::sub(const Value& a, const Value& b)
{
	return Value( a.i - b.toInt() );
}

Value IntegerPrimitive::neg(const Value& x)
{
	return Value( - x.i );
}

Value IntegerPrimitive::mul(const Value& a, const Value& b)
{
	return Value( a.i * b.toInt() );
}

Value IntegerPrimitive::div(const Value& a, const Value& b)
{
	int denom = b.toInt();
	if( denom == 0 ) error("divide by zero");
	return Value( a.i / denom);
}

Value IntegerPrimitive::mod(const Value& a, const Value& b)
{
	int denom = b.toInt();
	if( denom == 0 ) error("divide by zero");
	return Value( a.i % denom);
}

Value IntegerPrimitive::cmp(const Value& a, const Value& b)
{
	return Value( a.i - b.toInt() );
}

Value IntegerPrimitive::str(const Value& a)
{
	char buf[32];
	snprintf(buf,31,"%d", a.i);
	return Value( valueStr, MemeString::newString( buf ) );
}

Value IntegerPrimitive::num(const Value& a)
{
	return a;
}

