#ifndef __MEME__
#define __MEME__

// Bison/yacc stuff
#define YYERROR_VERBOSE 1

#define MEME_TRACE 0

#include "memetype.h"

typedef enum {
	typeCon=0,
	typeReal,
	typeNull,
	typeArray,
	typeBool,
	typeStr,
	typeId,
	typeOpr,
	typeFunc,
	typeCall,
	typeParam
} nodeEnum;

/* constants */
typedef struct {
	union {
		int value;                  // value of constant
		bool bval;
		double rval;
	};
} conNodeType;

/* strings */
typedef struct {
	MemeRef ref;				/* ref to gc'ed string */
} strNodeType;

/* identifiers */
typedef struct {
	const char *name; 				/* symbol name */
} idNodeType;

/* operators */
typedef struct {
    int oper;                   /* operator */
    int nops;                   /* number of operands */
    struct nodeTypeTag *op[1];  /* operands (expandable) */
} oprNodeType;

/* function */
typedef struct {
	const char *name;
    struct nodeTypeTag *params, *impl;
} funcNodeType;

/* function call */
typedef struct {
	const char *name; 				/* function name */
	struct nodeTypeTag *args;
} callNodeType;

/* param name */
typedef struct {
	const char *name;
} paramNodeType;

/* array node */
typedef struct {
	struct nodeTypeTag *elements;
} arrayNodeType;


typedef struct nodeTypeTag {
    nodeEnum type;              // type of node

	struct nodeTypeTag *next; // used for chaining arguments
	strNodeType str; // has a constructor, so hates union

    // union must be last entry in nodeType
    // because operNodeType may dynamically increase
    union {
        conNodeType con;        /* constants */
        idNodeType id;          /* identifiers */
        oprNodeType opr;        /* operators */
		funcNodeType func; 		/* functions */
		callNodeType call; 		/* call func */
		paramNodeType param; 	/* define param */
		arrayNodeType array;
    };
	unsigned char gaurd; // ugly hack

} nodeType;


void trace(const char *fmt, ...);
void warning(const char *fmt, ...);
void error(const char *fmt, ...);

void countLine(void);

nodeType *opr(int oper, int nops, ...);
nodeType *id(const char *name);
nodeType *param(const char *name);
nodeType *con(int value);
nodeType *newArray(nodeType *args);
nodeType *chainArg(nodeType *left, nodeType *right);
nodeType *nullval(void);
nodeType *boolval(bool bval);
nodeType *realval(double value);
nodeType *str(char *text);
nodeType *makeFunctionNode(const char *name, nodeType *params);
nodeType *define_function(const char *name, nodeType *params, nodeType *impl);
nodeType *callFunction(const char *name, nodeType *args);
void freeNode(nodeType *p);
class MemeScope;
Value ex(MemeScope& locals, nodeType *p);

void setup(int argc, const char **argv, char **envp);

char *storeString(const char *str);
char *adoptString(char *str);
char *stringAdd(const char *s1, const char *s2);
void freeStrings(void);


#endif

