#ifndef __MEMEBASE__
#define __MEMEBASE__

#include <list>
#include <stddef.h> // for NULL

// base type everything inherits from that will be deleted by something else.
class MemeBase {
	public:
		MemeBase();
		virtual ~MemeBase();
};

class MemeReaper;
class MemeRef;

// base type for everything in GC
class MemeCollect : public MemeBase {
	private:
		int refCount_;
		bool untracked_;
	public:
		MemeCollect(void);
		
		// if you call this before a ref is created to an object,
		// it lets you managed it and doesn't go to the GC/reaper
		void untracked( void ) { untracked_= true; } // tricks MemeRef into thinking on GC already
	protected:
		~MemeCollect();
	private:
		friend class MemeReaper;
		friend class MemeRef;

		int refCount(void);
		int incRefCount(void);
		void decRefCount(void);
};

// this is a ref to something in the GC
class MemeRef : public MemeBase {
	private:
		MemeCollect* target_;
		void newRef(MemeCollect *target_);
	public:

		MemeRef(MemeCollect* obj = NULL);
		MemeRef(const MemeRef& mr);
		~MemeRef();
	
		MemeCollect *target(void) const { return target_; }
		void setTarget(MemeCollect*obj);

		const MemeRef& operator = (const MemeRef& mr);
};

class MemeReaper : public MemeBase {
	private:
		static MemeReaper *sReaper;

		std::list<MemeCollect*> pile_;
	public:
		MemeReaper(void);
		~MemeReaper();

		static MemeReaper* Reaper(void) { return sReaper; }

		void track( MemeCollect *mc );
		void reap(void);

		// gc unit test
		static bool test(void);
};

#endif


