/*
 * real.cpp
 *
 *  Created on: Mar 1, 2009
 *      Author: ranton
 */

#include "memetype.h"
#include "meme.h"
#include <cstdio>

class RealPrimitive: public MemePrimitive {
public:
	// these are not abstract, they all throw
	// errors if called on an object that does
	// not implement
	RealPrimitive();
	~RealPrimitive();

	// NOTE: this should have variable arguments
	virtual Value ini(const Value& x);

	virtual Value add(const Value& a, const Value& b);
	virtual Value sub(const Value& a, const Value& b);
	virtual Value neg(const Value& x);

	virtual Value mul(const Value& a, const Value& b);
	virtual Value div(const Value& a, const Value& b);

	virtual Value cmp(const Value& a, const Value& b);

	virtual Value str(const Value& a);
	virtual Value num(const Value& a);

};

MemePrimitive *getRealPrimitive(void) {
	return new RealPrimitive();
}

// these are not abstract, they all throw
// errors if called on an object that does
// not implement
RealPrimitive::RealPrimitive() {
}

RealPrimitive::~RealPrimitive() {
}

// NOTE: this should have variable arguments
Value RealPrimitive::ini(const Value& x) {
	return x;
}

Value RealPrimitive::add(const Value& a, const Value& b) {
	return Value(a.r + b.toReal());
}

Value RealPrimitive::sub(const Value& a, const Value& b) {
	return Value(a.r - b.toReal());
}

Value RealPrimitive::neg(const Value& x) {
	return Value(-x.r);
}

Value RealPrimitive::mul(const Value& a, const Value& b) {
	return Value(a.r * b.toReal());
}

Value RealPrimitive::div(const Value& a, const Value& b) {
	double denom = b.toReal();
	if (denom == 0.0)
		error("divide by zero");
	return Value(a.r / denom);
}


Value RealPrimitive::cmp(const Value& a, const Value& b) {
	return Value(a.r - b.toReal());
}

Value RealPrimitive::str(const Value& a) {
	char buf[100];
	snprintf(buf, 100, "%f", a.r);
	return Value(valueStr, MemeString::newString(buf));
}

Value RealPrimitive::num(const Value& a) {
	return a;
}

