#include <string.h>
#include <stdlib.h>
#include <stdio.h>


#include "memetype.h"
#include "meme.h"

class StringPrimitive : public MemePrimitive {
	public:
		// these are not abstract, they all throw
		// errors if called on an object that does
		// not implement
		StringPrimitive();
		~StringPrimitive();

		// NOTE: this should have variable arguments
		virtual Value ini(const Value& x);

		virtual Value add(const Value& a, const Value& b);

		virtual Value cmp(const Value& a, const Value& b);
		// NOTE: need idx as well
		
		virtual Value str(const Value& a);
		virtual Value num(const Value& a);
		
		virtual Value len(const Value& a);
		virtual Value idx(const Value& a, const Value& i);
		virtual Value cat(const Value& a, const Value& b);
};

MemePrimitive *getStringPrimitive(void)
{
	return new StringPrimitive();
}

// these are not abstract, they all throw
// errors if called on an object that does
// not implement
StringPrimitive::StringPrimitive()
{
}

StringPrimitive::~StringPrimitive()
{
}


// NOTE: this should have variable arguments
Value StringPrimitive::ini(const Value& x)
{
	return x;
}


Value StringPrimitive::add(const Value& a, const Value& b)
{
	MemeString *result = new MemeString( MemeString::strFromRef( a.ref_ ) );
	result->append( MemeString::strFromRef( b.getPrimitive()->str(b).ref_ ) );
	return Value( valueStr, MemeRef( result ) );

}

Value StringPrimitive::cmp(const Value& a, const Value& b)
{
	return strcmp( MemeString::strFromRef(a.ref_), b.toString() );
}

Value StringPrimitive::str(const Value& a)
{
	return a;
}

Value StringPrimitive::num(const Value& a)
{
	// NOTE: deal with floating point possibilities as well
	return Value( atoi( MemeString::strFromRef( a.ref_ ) ) );
}

Value StringPrimitive::len(const Value& a)
{
	const char* str = MemeString::strFromRef( a.ref_ );
	int slen = (int)strlen( str );
	return Value(slen);
}

Value StringPrimitive::idx(const Value& a, const Value& i)
{
	// create and return a string with just that index.
	const char* str = MemeString::strFromRef( a.ref_ );
	char cbuf[2];
	int idx = i.toInt();
	if( idx >=0 && idx <= (int)strlen( str ) )
	{
		cbuf[0] = str[ idx ]; 
		cbuf[1] = 0;
	}
	else
		cbuf[0] = 0;
	MemeRef sref = MemeString::newString( cbuf ) ;
	return Value(valueStr, sref);
}

Value StringPrimitive::cat(const Value& a, const Value& b)
{
	MemeString *actual   = MemeString::derefString( a.ref_ );
	actual->append( MemeString::strFromRef( b.getPrimitive()->str(b).ref_ ) );
	return b;
}


/// NOTE: for now we are going to continue to use
// these for the parsing, but we are NOT
// going to use them for anything else.

typedef struct StringList_t {
	char *text;
	struct StringList_t *next;
} StringList;

static StringList *sStrings = NULL;
// store a string for use and garbage collection later
char *storeString(const char *str) {
	char *s = strdup( str );
	if( !s ) error("out of memory");
	return adoptString(s);
}

char *adoptString(char *str) {
	StringList *s = (StringList *)malloc(sizeof(StringList));
	if( !s ) error("out of memory");
	s->text = str;
	s->next = sStrings;
	sStrings = s;
	return s->text;
}

char *stringAdd(const char *s1, const char *s2)
{
	int len1, len2;
	len1 = strlen(s1);
	len2 = strlen(s2);
	char *s = (char *)malloc(sizeof(char)*(len1+len2+1));
	if( !s ) error("out of memory");
	strcpy(s,s1);
	strcpy(s+len1,s2);
	return adoptString( s );
}

// free all strings
void freeStrings(void) {
	StringList *n,*s = sStrings;
	while( s )
	{
		n = s->next;
		free((void *)s);
		s = n;
	}
}
