/*
 * typesupport.cpp
 * This contains basic builtin functions such as type() and len()
 *
 *  Created on: Mar 1, 2009
 *      Author: ranton
 */


#include "meme.h"
#include "memetype.h"
#include "memeutil.h"
#include "memearray.h"
#include "memehash.h"

#include <cstring>
#include <iostream>

using namespace std;



static Value bi_len( MemeScope& scope )
{
	Value a = getVarValue(scope,"a");
	return a.getPrimitive()->len(a);
}

static Value bi_type( MemeScope& scope )
{
	Value a = getVarValue(scope,"a");
	const char *name = "(unknown)";

	switch(a.type)
	{
	case valueNull:
		name = "null";
		break;
	case valueInt:
		name = "integer";
		break;
	case valueReal:
		name = "real";
		break;
	case valueBool:
		name = "bool";
		break;
	case valueStr:
		name = "string";
		break;
	case valueArray:
		name = "array";
		break;
	case valueHash:
		name = "hash";
		break;
	case valueRef:
		name = "ref";
		break;
	case valueAppender:
		name = "appender";
		break;
	case valueFunc:
		name = "function";
		break;
	}
	return Value( valueStr, MemeString::newString( name ) );
}
static Value bi_int( MemeScope& scope ) {
	Value a = getVar(scope,"a");
	return a.toInt();
}
static Value bi_real( MemeScope& scope ) {
	Value a = getVar(scope,"a");
	return a.toReal();
}
static Value bi_str( MemeScope& scope ) {
	Value a = getVar(scope,"a");
	return a.getPrimitive()->str( a );
}

static Value bi_substr( MemeScope& scope ) {
	int i = getVar(scope,"i").toInt();
	int len = getVar(scope,"len").toInt();
	const char *str = getVar(scope,"s").toString();
	while( *str && i )
	{
		str++;
		i--;
	}
	if( i )
		return Value(); // ran out of string
	int use = 0;
	while( str[use] && use < len )
		use++; // Inefficient, but we don't' want it to explode.
	return Value( valueStr, MemeString::newString( str, use ) );
}

static Value bi_char( MemeScope& scope ) {
	int i = getVar(scope,"i").toInt();
	char s[2];
	s[0] = (char)i;
	s[1] = 0;
	return Value( valueStr, MemeString::newString( s, 2) );
}
static Value bi_ord( MemeScope& scope) {
	const char *str = getVar(scope,"s").toString();
	int i = *str; // if empty will just get 0 anyway
	return Value( i );
}
static Value bi_range( MemeScope& scope ) {
	// NOTE: we should really just return a special type of object
	// that acts as a collection without storing this crap
	// !!!
	int lo = getVar(scope,"lo").toInt();
	int hi = getVar(scope,"hi").toInt();
	// create an arrow from [lo,hi]
	Value arr = Value(valueArray,MemeArray::newArray());

	if( lo < hi ) {
		for(int x = lo; x <= hi; x++ )
			arr.getPrimitive()->cat( arr, Value(x) );
	} else {
		for(int x = lo; x >= hi; x-- )
			arr.getPrimitive()->cat( arr, Value(x) );
	}
	return arr;
}
static Value bi_hash( MemeScope& scope ) {
	return Value( valueHash, MemeHash::newHash());
}

static Value bi_keys( MemeScope& scope ) {
	Value h = getVarValue(scope,"h");
	Value result;
	if( h.type == valueHash )
		result = ((MemeHash*)h.ref_.target())->keys();
	return result;
}
static Value bi_values( MemeScope& scope ) {
	Value h = getVarValue(scope,"h");
	Value result;
	if( h.type == valueHash )
		result = ((MemeHash*)h.ref_.target())->values();
	return result;
}
static Value bi_exists( MemeScope& scope ) {
	Value h = getVarValue(scope,"h");
	Value k = getVar(scope,"k");
	Value result;
	if( h.type == valueHash )
		result = Value( ((MemeHash*)h.ref_.target())->exists( k ) );
	return result;
}
static Value bi_erase( MemeScope& scope ) {
	Value h = getVarValue(scope,"h");
	Value k = getVar(scope,"k");
	Value result;
	if( h.type == valueHash )
		 result = ((MemeHash*)h.ref_.target())->erase( k );
	return result;
}

static Value bi_join(MemeScope& scope) {
#if MEME_TRACE
	cout << "bi_join" << endl;
#endif
	// delim as string value
	Value delim = getVarValue(scope,"d");
#if MEME_TRACE
	delim.println();
#endif

	// array and length
	Value a = getVarValue(scope,"a");
	int alen = a.getPrimitive()->len(a).toInt();
#if MEME_TRACE
	a.println();
	std::cout << "alen = " << alen << std::endl;
#endif

	// empty result string
	Value result = Value( valueStr, MemeString::newString( "" ) );

	for(int i=0;i<alen;i++) {
		// get value of array index
		Value v = a.getPrimitive()->idx(a,Value(i));
		// add delim
		if(i>0)
			result.getPrimitive()->cat(result,delim);
		// add value
		result.getPrimitive()->cat(result,v); // s or v TODO:
	}

	return result;
}

static Value bi_split(MemeScope& scope) {
	const char *delim = getVar(scope,"d").toString();
	const char *str = getVar(scope,"s").toString();
	int dlen = strlen(delim);
	// int slen = strlen(str);

	MemeArray *arr = new MemeArray();

	const char *next, *last = str;

	if(dlen>0) {
		while( (next=strstr(last,delim)) != NULL )
		{ // have delim if here
			arr->push( Value( valueStr, MemeString::newString( last, next-last ) ) );
			last = next + dlen;
		}
	}
	// no delim
	arr->push( Value( valueStr, MemeString::newString( last ) ) );

	return Value(valueArray, MemeRef(arr));

}
void RegisterTypeSupport(MemeScope& scope)
{
	RegisterCppFunc(scope, "len", "a", bi_len );
	RegisterCppFunc(scope, "type", "a", bi_type );
	RegisterCppFunc(scope, "int", "a", bi_int );
	RegisterCppFunc(scope, "real", "a", bi_real );
	RegisterCppFunc(scope, "str", "a", bi_str );
	RegisterCppFunc(scope, "substr", "s,i,len", bi_substr);
	RegisterCppFunc(scope, "char", "i", bi_char);
	RegisterCppFunc(scope, "ord", "s", bi_ord);
	RegisterCppFunc(scope, "range", "lo,hi", bi_range);
	RegisterCppFunc(scope, "hash", "", bi_hash);
	RegisterCppFunc(scope, "keys", "h", bi_keys);
	RegisterCppFunc(scope, "values", "h", bi_values);
	RegisterCppFunc(scope, "exists", "h,k", bi_exists);
	RegisterCppFunc(scope, "erase", "h,k", bi_erase);
	RegisterCppFunc(scope, "join", "d,a", bi_join);
	RegisterCppFunc(scope, "split", "d,s", bi_split);

	RegisterConstant(scope, "EOL", Value( valueStr, MemeString::newString( "\n" ) ) );
	RegisterConstant(scope, "CR", Value( valueStr, MemeString::newString( "\r" ) ) );
	RegisterConstant(scope, "TAB", Value( valueStr, MemeString::newString( "\t" ) ) );
}
